<p align="center">
<img src='https://github.com/aaspip/gallery/blob/main/aaspip/aaspip-logo.png' alt='comp' width=320/>
</p>

**aaspip**
======

## Description

**aaspip** is a python package for structural denoising and interpolation of multi-channel seismic data. The latest version has incorporated both Python and C (hundreds of times faster) implementations of the embedded functions. We keep both implementations for both educational and production purposes. This package has a variety of applications in both exploration and earthquake seismology.

## Reference
Chen et al., 2023, AASPIP: Advanced Array Seismic data Processing and Imaging Platform, TBD. 

BibTeX:

	@article{aaspip,
	  title={AASPIP: Advanced Array Seismic data Processing and Imaging Platform},
	  author={Yangkang Chen },
	  journal={TBD},
	  volume={TBD},
	  number={TBD},
	  pages={TBD},
	  year={TBD}
	}

-----------
## Copyright
    The aaspip developing team, 2021-present
    
-----------
## License
    MIC License  

-----------
## Install
Using the latest version

    git clone https://github.com/aaspip/aaspip
    cd aaspip
    pip install -v -e .
or using Pypi

    pip install aaspip
    
or using pip directly from github (Preferrable because aaspip is updated very frequently)

    pip install git+https://github.com/aaspip/aaspip

-----------
## Verified runnable OS
Mac OS, Linux, Windows (need Microsoft C++ Build Tools)

-----------
## DEMO scripts
    The "demo" directory contains all runable scripts to demonstrate different applications of aaspip. 

-----------
## Gallery
The gallery figures of the aaspip package can be found at
    https://github.com/aaspip/gallery/tree/main/aaspip
Each figure in the gallery directory corresponds to a DEMO script in the "demo" directory with the exactly the same file name.

-----------
## Dependence Packages
* scipy 
* numpy 
* matplotlib

-----------
## Modules
    dip2d.py  	-> 2D local slope estimation (including both python and C implementations)
    dip3d.py  	-> 3D local slope estimation (including both python and C implementations)
    divne.py  	-> element-wise division constrained by shaping regularization (python implementation)
    somean2d.py 	-> 2D structure-oriented mean filter  (including both python and C implementations)
    somean3d.py 	-> 3D structure-oriented mean filter  (including both python and C implementations)
    somf2d.py 	-> 2D structure-oriented median filter  (including both python and C implementations)
    somf3d.py 	-> 3D structure-oriented median filter  (including both python and C implementations)
    soint2d.py  	-> 2D structural interpolation  (including both python and C implementations)
    soint3d.py  	-> 3D structural interpolation  (including both python and C implementations)
    ricker.py	-> Ricker wavelet
    bp.py		-> Butterworth bandpass filter (including both python and C implementations)
    fk.py		-> FK dip filter
    plot.py		-> seismic plotting functions
    
-----------
## Development
    The development team welcomes voluntary contributions from any open-source enthusiast. 
    If you want to make contribution to this project, feel free to contact the development team. 

-----------
## Contact
    Regarding any questions, bugs, developments, collaborations, please contact  
    Yangkang Chen
    chenyk2016@gmail.com

-----------
## Examples
# Example 1 (2D structure-oriented mean/smoothing filter) 
Generated by [demos/test_aaspip_seistr_somean2d.py](https://github.com/aaspip/aaspip/tree/main/demos/test_aaspip_seistr_somean2d.py)

<img src='https://github.com/aaspip/gallery/blob/main/pyseistr/test_pyseistr_somean2d.png' alt='Slicing' width=960/>

# Example 2 (3D structure-oriented mean/smoothing filter) 
Generated by [demos/test_aaspip_seistr_somean3d.py](https://github.com/aaspip/aaspip/tree/main/demos/test_aaspip_seistr_somean3d.py)

<img src='https://github.com/aaspip/gallery/blob/main/pyseistr/test_pyseistr_somean3d.png' alt='Slicing' width=960/>

# Example 3 (2D structure-oriented median filter) 
Generated by [demos/test_aaspip_seistr_somf2d.py](https://github.com/aaspip/aaspip/tree/main/demos/test_aaspip_seistr_somf2d.py)

<img src='https://github.com/aaspip/gallery/blob/main/pyseistr/test_pyseistr_somf2d.png' alt='Slicing' width=960/>

# Example 4 (3D structure-oriented median filter) 
Generated by [demos/test_aaspip_seistr_somf3d.py](https://github.com/aaspip/aaspip/tree/main/demos/test_aaspip_seistr_somf3d.py)

<img src='https://github.com/aaspip/gallery/blob/main/pyseistr/test_pyseistr_somf3d.png' alt='Slicing' width=960/>

# Example 5 (3D structure-oriented interpolation) 
Generated by [demos/test_aaspip_seistr_passive_recon3d.py](https://github.com/aaspip/aaspip/tree/main/demos/test_aaspip_seistr_passive_recon3d.py)

<img src='https://github.com/aaspip/gallery/blob/main/pyseistr/test_pyseistr_passive_recon3d.png' alt='Slicing' width=960/>

# Example 6 (SS precursor data enhancement) 
Generated by [demos/test_aaspip_seistr_ssprecursor.py](https://github.com/aaspip/aaspip/tree/main/demos/test_aaspip_seistr_ssprecursor.py)

<img src='https://github.com/aaspip/gallery/blob/main/pyseistr/test_pyseistr_ssprecursor.png' alt='Slicing' width=960/>

# Example 7 (receiver function data enhancement) 
Generated by [demos/test_aaspip_seistr_rf.py](https://github.com/aaspip/aaspip/tree/main/demos/test_aaspip_seistr_rf.py)

<img src='https://github.com/aaspip/gallery/blob/main/pyseistr/test_pyseistr_rf.png' alt='Slicing' width=960/>

# Example 8 (structure-oriented distributed acoustic sensing (DAS) data processing) 
Generated by [demos/test_aaspip_seistr_das.py](https://github.com/aaspip/aaspip/tree/main/demos/test_aaspip_seistr_das.py)

<img src='https://github.com/aaspip/gallery/blob/main/pyseistr/test_pyseistr_das.png' alt='Slicing' width=960/>

# Below are new examples in addition to the results in the original paper

# Below is an example for 2D structure-oriented interpolation of a multi-channel synthetic seismic data
Generated by [demos/test_aaspip_seistr_soint2d.py](https://github.com/aaspip/aaspip/tree/main/demos/test_aaspip_seistr_soint2d.py)
<img src='https://github.com/aaspip/gallery/blob/main/pyseistr/test_pyseistr_soint2d.png' alt='comp' width=960/>

# Below is an example for 2D structure-oriented interpolation of a seafloor dataset
Generated by [demos/test_aaspip_seistr_soint2d_seafloor.py](https://github.com/aaspip/aaspip/tree/main/demos/test_aaspip_seistr_soint2d_seafloor.py)
<img src='https://github.com/aaspip/gallery/blob/main/pyseistr/test_pyseistr_soint2d_seafloor.png' alt='comp' width=960/>

# Below is another example for 2D structure-oriented interpolation of sparse seismic data
Generated by [demos/test_aaspip_seistr_soint2d2.py](https://github.com/aaspip/aaspip/tree/main/demos/test_aaspip_seistr_soint2d2.py)
<img src='https://github.com/aaspip/gallery/blob/main/pyseistr/test_pyseistr_soint2dnew.png' alt='comp' width=960/>

# Below is an example for calculating relative geological time (RGT) from 2D seismic data
Generated by [demos/test_aaspip_seistr_rgt2d.py](https://github.com/aaspip/aaspip/tree/main/demos/test_aaspip_seistr_rgt2d.py)
<img src='https://github.com/aaspip/gallery/blob/main/pyseistr/test_pyseistr_rgt2d.png' alt='comp' width=960/>

# Below is an example for 2D interpolation of sparse seismic data
Generated by [demos/test_aaspip_seistr_sint2d.py](https://github.com/aaspip/aaspip/tree/main/demos/test_aaspip_seistr_sint2d.py)
<img src='https://github.com/aaspip/gallery/blob/main/pyseistr/test_pyseistr_sint2d.png' alt='comp' width=960/>

# Below is an example for plotting fancy 3D figures using aaspip
Generated by [demos/test_first](https://github.com/chenyk1990/pywave/tree/main/demos/test_first.py)
<p align="center">
<img src='https://github.com/chenyk1990/gallery/blob/main/pywave/vel3d.png' alt='comp' width=960/>
<img src='https://github.com/chenyk1990/gallery/blob/main/pywave/data3d.png' alt='comp' width=640/>
</p>

# Below is an example for generating fancy GIF figures using aaspip
Generated by [demos/test_second_wfd3ds](https://github.com/chenyk1990/pywave/tree/main/demos/test_second_wfd3ds.py)

<p align="center">
<img src='https://github.com/chenyk1990/gallery/blob/main/pywave/wfd3ds.gif' alt='comp' width=960/>
</p>

# Below is an example for generating fancy wiping-style animation using aaspip
Generated by [demos/test_aaspip_seistr_das_animation](https://github.com/aaspip/aaspip/tree/main/demos/test_aaspip_seistr_das_animation.py)

<p align="center">
<img src='https://github.com/aaspip/gallery/blob/main/pyseistr/test_pyseistr_das_animation.gif' alt='comp' width=960/>
</p>

# Below is an example for visualizing 3D patches
Generated by [mlnb/DL_denoise_simple3D.ipynb](https://github.com/chenyk1990/mlnb/blob/main/DL_denoise_simple3D.ipynb)

<p align="center">
<img src='https://github.com/aaspip/gallery/blob/main/aaspip/patches.png' alt='comp' width=960/>
</p>


## Examples of EKFMM

DEMO1 
The following figure shows an example of traveltime calculation for 2D isotropic media (a) and anisotropic media (b). Generated by [demos/test_aaspip_ekfmm_fig1.py](https://github.com/aaspip/aaspip/blob/main/demos/test_aaspip_ekfmm_fig1.py)
<img src='https://github.com/aaspip/gallery/blob/main/pyekfmm/test_pyekfmm_fig1.png' alt='DEMO1' width=960/>

DEMO2
The following figure shows an example of traveltime calculation for 3D isotropic media (a) and anisotropic media (b). Generated by [demos/test_aaspip_ekfmm_fig2.py](https://github.com/aaspip/aaspip/blob/main/demos/test_aaspip_ekfmm_fig2.py)
<img src='https://github.com/aaspip/gallery/blob/main/pyekfmm/test_pyekfmm_fig2.png' alt='DEMO2' width=960/>

